/* Copyright (c) 2012 Oberon microsystems, Inc. (Switzerland)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. */

// Originally developed for the book
//   "Getting Started with the Internet of Things", by Cuno Pfister.
//   Copyright 2011 Cuno Pfister, Inc., 978-1-4493-9357-1.

// Version 1.0, for the .NET Micro Framework release 4.2.

using System.Diagnostics.Contracts;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;

namespace Gsiot.Http.Ethernet
{
    // server-side: listener

    public sealed class EthernetStreamListener : IStreamListener
    {
        public bool IsOpen { get; private set; }
        public string LocalHostName { get; private set; }
        public int LocalPort { get; set; }

        // configuration properties specific to sockets (usually Ethernet sockets)
        public int Backlog { get; set; }

        Socket listener;

        public EthernetStreamListener()
        {
            IsOpen = false;
            LocalHostName = null;
            LocalPort = -1;
            Backlog = 0;
        }

        public void Open()
        {
            Contract.Requires(!IsOpen);
            Contract.Requires(LocalPort >= 0);
            Contract.Requires(LocalPort <= 65535);
            if (Backlog <= 0) { Backlog = 4; }      // TODO good default?

            Contract.Assert(listener == null);
            try
            {
                IPAddress myAdr = IPAddress.GetDefaultLocalAddress();
                while (myAdr == IPAddress.Any)      // only necessary for lwIP stack?
                {
                    Thread.Sleep(1000);             // wait until local address is set
                    myAdr = IPAddress.GetDefaultLocalAddress();
                }
                LocalHostName = myAdr.ToString();
                listener = new Socket(AddressFamily.InterNetwork,
                                        SocketType.Stream, ProtocolType.Tcp);
                listener.SetSocketOption(SocketOptionLevel.Tcp, SocketOptionName.NoDelay, true);
                listener.Bind(new IPEndPoint(IPAddress.Any, LocalPort));
                listener.Listen(Backlog);
                IsOpen = true;
            }
            catch (SocketException e)
            {
                if (listener != null) { Dispose(); }
                listener = null;
                throw new StreamException("socket error " + e.ErrorCode, e);
            }
        }

        public void Dispose()
        {
            if (IsOpen)
            {
                IsOpen = false;
                Contract.Assert(listener != null);
                listener.Close();
                listener = null;
                LocalHostName = null;
            }
        }

        public void Close()
        {
            Dispose();
        }

        public Stream Accept()
        {
            try
            {
                Socket s = listener.Accept();
                // Socket.Accept may throw an exception: SocketException,
                // ObjectDisposedException, or InvalidOperationException
                return new NetworkStream(s);
            }
            catch (SocketException e)     // this may happen, treat as expected error
            {
                throw new StreamException("socket error " + e.ErrorCode, e);
            }
        }
    }


    // client-side: socket

    public sealed class EthernetStreamFactory : IStreamFactory
    {
        public bool IsOpen { get; private set; }
        public string RemoteHostName { get; set; }
        public int RemotePort { get; set; }

        public EthernetStreamFactory()
        {
            IsOpen = false;
            RemoteHostName = null;
            RemotePort = -1;
        }

        public void Open()
        {
            Contract.Requires(!IsOpen);
            Contract.Requires(RemoteHostName != null);
            Contract.Requires(RemoteHostName.Length > 0);
            Contract.Requires(RemotePort >= 0);
            Contract.Requires(RemotePort <= 65535);

            IsOpen = true;
        }

        public void Dispose()
        {
            if (IsOpen)
            {
                IsOpen = false;
                RemoteHostName = null;
            }
        }

        public void Close()
        {
            Dispose();
        }

        public Stream Connect()
        {
            Contract.Requires(IsOpen);

            // look up relay host's domain name,
            // to find IP address(es)
            IPHostEntry hostEntry = Dns.GetHostEntry(RemoteHostName);
            // SocketException in GetHostEntry would
            // indicate that host was not found
            //Debug.Print(host + " is resolved");
            // extract a returned address
            try
            {
                IPAddress hostAddress = hostEntry.AddressList[0];
                IPEndPoint remoteEndPoint = new IPEndPoint(hostAddress, RemotePort);

                var socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                socket.SetSocketOption(SocketOptionLevel.Tcp, SocketOptionName.NoDelay, true);
                socket.Connect(remoteEndPoint);
                return new NetworkStream(socket);
            }
            catch (SocketException s)
            {
                Dispose();
                throw new StreamException("socket exception " + s.ErrorCode, s);
            }
        }
    }
}
